module FileSystem

open util/ordering[FileSystem] as Ord

abstract sig Object {}

sig File, Dir extends Object {}

sig FileSystem {
	objects : set Object,
	contents : Dir -> Object
}

pred Contents [fs : FileSystem] {
/*
	univ.(fs.contents) in fs.objects
	(fs.contents).univ in fs.objects
*/
	fs.contents in fs.objects -> fs.objects
}

pred OneParent [fs : FileSystem] {
	all f : File & fs.objects | one fs.contents.f 
	all d : Dir & fs.objects | lone fs.contents.d
	one d : Dir & fs.objects | no fs.contents.d
}

pred Acyclic [fs : FileSystem] {
	no ^(fs.contents) & iden
}

pred Connected [fs : FileSystem] {
	some d : Dir & fs.objects | fs.objects in (d.^(fs.contents) + d)
}

pred Inv [fs : FileSystem] {
	Contents[fs] && OneParent[fs] && Acyclic[fs]
}

run Inv for 5 but exactly 1 FileSystem

check {all fs : FileSystem | Inv[fs] => Connected[fs]} for 5

// make

pred make [fs , fs' : FileSystem, d : Dir, o : Object] {
	// PRE
	d in fs.objects
	o !in fs.objects
	// POS
	fs'.objects = fs.objects + o
	fs'.contents = fs.contents + d->o
}

assert makeOk {
	all fs, fs' : FileSystem, d : Dir, o : Object {
		Inv[fs] && make[fs,fs',d,o] => Inv[fs']
	}	
}

check makeOk for 5 but 2 FileSystem

pred makeTest [fs , fs' : FileSystem, d : Dir, o : Object] {
	Inv[fs] && make[fs,fs',d,o]
}

run makeTest for 4 but 2 FileSystem

// rm

pred rm [fs, fs' : FileSystem, f : File] {
	// PRE
	f in fs.objects
	// POS
	fs'.objects = fs.objects - f
	// fs'.contents = fs.contents :> (univ - f)
	fs'.contents = fs.contents - (Dir -> f)
}

assert rmOk {
	all fs, fs' : FileSystem, f : File {
		Inv[fs] && rm[fs,fs',f] => Inv[fs']
	}	
}

check rmOk for 5 but 2 FileSystem

pred rmTest [fs , fs' : FileSystem, f : File] {
	Inv[fs] && rm[fs,fs',f]
}

run rmTest for 4 but 2 FileSystem

// rmdir

pred rmdir [fs, fs' : FileSystem, d : Dir] {
	// PRE
	d in fs.objects             // d tem que existir
	no d.(fs.contents)        // d tem que ser vazia
	some (fs.contents).d    // d não pode ser a raiz
	// POS
	fs'.objects = fs.objects - d
	fs'.contents = fs.contents - (Dir -> d)
}

assert rmdirOk {
	all fs, fs' : FileSystem, d : Dir {
		Inv[fs] && rmdir[fs,fs',d] => Inv[fs']
	}	
}

check rmdirOk for 5 but 2 FileSystem

pred rmdirTest [fs , fs' : FileSystem, d : Dir] {
	Inv[fs] && rmdir[fs,fs',d]
}

run rmdirTest for 4 but 2 FileSystem

// rmrf

fun root [fs : FileSystem] : set Dir {
	{d : Dir & fs.objects | no fs.contents.d}
}

check {all fs : FileSystem | Inv[fs] => one root[fs]} for 5

pred rmrf [fs, fs' : FileSystem, o : Object] {
	// PRE
	o in fs.objects
	// POS
	fs'.objects = fs.objects - o.*(fs.contents) + root[fs]
	fs'.contents = fs.contents - (Dir -> o.*(fs.contents))
}

assert rmrfOk {
	all fs, fs' : FileSystem, o : Object {
		Inv[fs] && rmrf[fs,fs',o] => Inv[fs']
	}	
}

check rmrfOk for 5 but 2 FileSystem

pred rmrfTest [fs , fs' : FileSystem, o : Object] {
	Inv[fs] && rmrf[fs,fs',o]
}

run rmrfTest for 4 but 2 FileSystem

// mv

pred mv [fs, fs' : FileSystem, o : Object, d : Dir] {
	// PRE
	(d+o) in fs.objects
	d !in o.*(fs.contents) // Nao e possivel mover para um descendente
	// POS
	fs'.objects = fs.objects
	fs'.contents = fs.contents - (Dir -> o) + (d -> o)
}

assert mvOk {
	all fs, fs' : FileSystem, o : Object, d : Dir {
		Inv[fs] && mv[fs,fs',o,d] => Inv[fs']
	}	
}

check mvOk for 5 but 2 FileSystem

pred mvTest [fs , fs' : FileSystem, o : Object, d : Dir] {
	Inv[fs] && mv[fs,fs',o,d]
}

run mvTest for 4 but 2 FileSystem

// Ordering

fact First {
	Inv[first]
}

fact Next {
	all fs : FileSystem, fs' : next[fs] {
		some o : Object, d : Dir, f : File {
			make[fs,fs',d,o] || rm[fs,fs',f] || rmdir[fs,fs',d] || rmrf[fs,fs',o] || mv[fs,fs',o,d]
		}
	}	
}

run {} for 4 but 7 FileSystem

check {all fs : FileSystem | Inv[fs]} for 5 but 7 FileSystem


